/*
 * Name:    d_cmptrans.c 
 *
 * Purpose: MX motor driver to move 1 or more motors in the same Compumotor
 *          controller in the same direction.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_motor.h"
#include "i_cmpmtr.h"
#include "d_cmpmtr.h"
#include "d_cmptrans.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_compumotor_trans_record_function_list = {
	mxd_compumotor_trans_initialize_type,
	mxd_compumotor_trans_create_record_structures,
	mxd_compumotor_trans_finish_record_initialization,
	mxd_compumotor_trans_delete_record,
	mxd_compumotor_trans_print_motor_structure,
	mxd_compumotor_trans_read_parms_from_hardware,
	mxd_compumotor_trans_write_parms_to_hardware,
	mxd_compumotor_trans_open,
	mxd_compumotor_trans_close
};

MX_MOTOR_FUNCTION_LIST mxd_compumotor_trans_motor_function_list = {
	mxd_compumotor_trans_motor_is_busy,
	mxd_compumotor_trans_move_absolute,
	mxd_compumotor_trans_get_position,
	mxd_compumotor_trans_set_position,
	mxd_compumotor_trans_soft_abort,
	mxd_compumotor_trans_immediate_abort,
	mxd_compumotor_trans_positive_limit_hit,
	mxd_compumotor_trans_negative_limit_hit,
	mxd_compumotor_trans_find_home_position,
	mxd_compumotor_trans_constant_velocity_move,
	mxd_compumotor_trans_get_parameter,
	mxd_compumotor_trans_set_parameter
};

/* Translation motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_compumotor_trans_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_COMPUMOTOR_TRANS_MOTOR_STANDARD_FIELDS
};

long mxd_compumotor_trans_num_record_fields
		= sizeof( mxd_compumotor_trans_record_field_defaults )
			/ sizeof( mxd_compumotor_trans_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_compumotor_trans_rfield_def_ptr
			= &mxd_compumotor_trans_record_field_defaults[0];

/* === */

#define NUM_COMPUMOTOR_TRANS_MOTOR_FIELDS	1

/* A private function for the use of the driver. */

static mx_status_type
mxd_compumotor_trans_get_pointers( MX_MOTOR *motor,
		MX_COMPUMOTOR_TRANSLATION_MOTOR **compumotor_translation_motor,
		const char *calling_fname )
{
	const char fname[] = "mxd_compumotor_trans_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( compumotor_translation_motor
		== (MX_COMPUMOTOR_TRANSLATION_MOTOR **) NULL )
	{
		return mx_error( MXE_NULL_ARGUMENT, fname,
	"The MX_COMPUMOTOR_TRANSLATION_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*compumotor_translation_motor = (MX_COMPUMOTOR_TRANSLATION_MOTOR *)
					motor->record->record_type_struct;

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_compumotor_trans_initialize_type( long type )
{
        const char fname[] = "mxs_compumotor_trans_motor_initialize_type()";

        const char field_name[NUM_COMPUMOTOR_TRANS_MOTOR_FIELDS]
		[MXU_FIELD_NAME_LENGTH+1]
            = {
                "motor_record_array"};

        MX_DRIVER *driver;
        MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
        MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
        MX_RECORD_FIELD_DEFAULTS *field;
        int i;
        long num_record_fields;
	long referenced_field_index;
        long num_motors_varargs_cookie;
        mx_status_type status;

        driver = mx_get_driver_by_type( mx_type_list, type );

        if ( driver == (MX_DRIVER *) NULL ) {
                return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
                        "Record type %ld not found in mx_type_list.", type );
        }

        record_field_defaults_ptr = driver->record_field_defaults_ptr;

        if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
                return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
                "'record_field_defaults_ptr' for record type '%s' is NULL.",
                        driver->name );
        }

        record_field_defaults = *record_field_defaults_ptr;

        if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
                return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
                "'record_field_defaults_ptr' for record type '%s' is NULL.",
                        driver->name );
        }

        if ( driver->num_record_fields == (long *) NULL ) {
                return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
                "'num_record_fields' pointer for record type '%s' is NULL.",
                        driver->name );
        }

	num_record_fields = *(driver->num_record_fields);

        status = mx_find_record_field_defaults_index(
                        record_field_defaults, num_record_fields,
                        "num_motors", &referenced_field_index );

        if ( status.code != MXE_SUCCESS )
                return status;

        status = mx_construct_varargs_cookie(
                        referenced_field_index, 0, &num_motors_varargs_cookie);

        if ( status.code != MXE_SUCCESS )
                return status;

        MX_DEBUG( 2,("%s: num_motors varargs cookie = %ld",
                        fname, num_motors_varargs_cookie));

	for ( i = 0; i < NUM_COMPUMOTOR_TRANS_MOTOR_FIELDS; i++ ) {
		status = mx_find_record_field_defaults(
			record_field_defaults, num_record_fields,
			field_name[i], &field );

		if ( status.code != MXE_SUCCESS )
			return status;

		field->dimension[0] = num_motors_varargs_cookie;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_trans_create_record_structures()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_trans_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	compumotor_trans_motor = (MX_COMPUMOTOR_TRANSLATION_MOTOR *)
			malloc( sizeof(MX_COMPUMOTOR_TRANSLATION_MOTOR) );

	if (compumotor_trans_motor == (MX_COMPUMOTOR_TRANSLATION_MOTOR *) NULL)
	{
		return mx_error( MXE_OUT_OF_MEMORY, fname,
    "Can't allocate memory for MX_COMPUMOTOR_TRANSLATION_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = compumotor_trans_motor;
	record->class_specific_function_list
				= &mxd_compumotor_trans_motor_function_list;

	motor->record = record;

	/* A compumotor_translation motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] =
		"mxd_compumotor_trans_finish_record_initialization()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	MX_MOTOR *motor;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	MX_RECORD *compumotor_interface_record;
	MX_RECORD *first_compumotor_interface_record;
	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	long i, num_motors, axis_number;
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	motor = (MX_MOTOR *) (record->record_class_struct);

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;

	/* Create the motor position and axis number arrays. */

	if ( num_motors <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		    "There must be at least one Compumotor motor record "
		    "specified for a Compumotor linear interpolation record.");
	} else {
		compumotor_translation_motor->motor_position_array = (double *)
				malloc( num_motors * sizeof(double) );
	
		if (compumotor_translation_motor->motor_position_array == NULL)
		{
			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Cannot allocate motor_position_array for %ld motors.",
			num_motors );
		}

		compumotor_translation_motor->index_to_axis_number = (int *)
				malloc( num_motors * sizeof(int) );
	
		if (compumotor_translation_motor->index_to_axis_number == NULL)
		{
			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Cannot allocate axis_number for %ld motors.",
			num_motors );
		}
	}

	/* Verify that only Compumotor motor records are in the motor
	 * record array.  Also initialize the recorded motor positions
	 * to 0.
	 */

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		if ( (child_motor_record->mx_superclass != MXR_DEVICE)
		  || (child_motor_record->mx_class != MXC_MOTOR)
		  || (child_motor_record->mx_type != MXT_MTR_COMPUMOTOR) )
		{
			return mx_error( MXE_TYPE_MISMATCH, fname,
			"Record '%s' is not a Compumotor motor record.",
				child_motor_record->name );
		}

		compumotor_translation_motor->motor_position_array[i] = 0.0;
	}

	/* Verify that the Compumotor motors are all controlled via the
	 * same Compumotor interface and record their axis numbers in
	 * local arrays.
	 */

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {
		compumotor_translation_motor->axis_number_to_index[i] = -1;
	}

	first_compumotor_interface_record = NULL;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		compumotor = (MX_COMPUMOTOR *)
				child_motor_record->record_type_struct;

		if ( i == 0 ) {
			first_compumotor_interface_record
				= compumotor->compumotor_interface_record;

			compumotor_translation_motor->controller_number
				= compumotor->controller_number;
		} else {
			compumotor_interface_record
				= compumotor->compumotor_interface_record;

			if ( compumotor_interface_record
			  != first_compumotor_interface_record )
			{
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"All motors in a Compumotor linear interpolation record must "
	"be controlled by the same Compumotor controller, but '%s' and '%s' "
	"are controlled by different controllers.",
					motor_record_array[0]->name,
					child_motor_record->name );
			}
		}

		axis_number = compumotor->axis_number;

		if ( (axis_number < 1)
		  || (axis_number > MX_MAX_COMPUMOTOR_AXES) )
		{
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Illegal Compumotor axis number %ld for motor '%s'.",
				axis_number, motor_record_array[i]->name );
		}

		compumotor_translation_motor->index_to_axis_number[i]
							= axis_number;

		/* We are storing 1-based indices in a C 0-based array,
		 * so we must add and subtract appropriate offsets of 1.
		 */

		compumotor_translation_motor->axis_number_to_index[
						axis_number - 1 ] = i + 1;
	}

	compumotor_translation_motor->compumotor_interface_record
		= first_compumotor_interface_record;

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			first_compumotor_interface_record->record_type_struct;

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_COMPUMOTOR_INTERFACE pointer for Compumotor interface '%s' is NULL.",
			first_compumotor_interface_record->name );
	}

	i = compumotor_translation_motor->controller_number - 1;

	compumotor_translation_motor->num_axes
				= compumotor_interface->num_axes[i];

#if 0
	fprintf(stderr, "\nindex_to_axis_number = ");

	for ( i = 0; i < compumotor_translation_motor->num_motors; i++ ) {
		fprintf( stderr, "%d ",
			compumotor_translation_motor->index_to_axis_number[i] );
	}
	fprintf(stderr, "\naxis_number_to_index = ");

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {
		fprintf( stderr, "%d ",
			compumotor_translation_motor->axis_number_to_index[i] );
	}
	fprintf(stderr, "\n");
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_delete_record( MX_RECORD *record )
{
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;

	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		compumotor_translation_motor
			= (MX_COMPUMOTOR_TRANSLATION_MOTOR *)
				record->record_type_struct;

		if (compumotor_translation_motor->motor_position_array != NULL)
		    free( compumotor_translation_motor->motor_position_array );

		if (compumotor_translation_motor->index_to_axis_number != NULL)
		    free( compumotor_translation_motor->index_to_axis_number );

		free( compumotor_translation_motor );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_print_motor_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_trans_print_motor_structure()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	MX_RECORD **motor_record_array;
	long i, num_motors;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_motors          = compumotor_translation_motor->num_motors;
	motor_record_array  = compumotor_translation_motor->motor_record_array;

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type           = COMPUMOTOR_TRANSLATION_MOTOR.\n\n");

	fprintf(file, "  name                 = %s\n", record->name);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position             = %g %s  (%g).\n",
		motor->position, motor->units, motor->raw_position.analog );
	fprintf(file, "  scale                = %g %s per raw unit.\n",
		motor->scale, motor->units);
	fprintf(file, "  offset               = %g %s.\n",
		motor->offset, motor->units);
	fprintf(file, "  backlash             = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog);
	fprintf(file, "  negative limit       = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog);
	fprintf(file, "  positive limit       = %g %s  (%g).\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.analog);

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband             = %g %s  (%g).\n\n",
		deadband, motor->units,
		motor->deadband.analog );

	fprintf(file, "  num motors           = %ld\n", num_motors);

	fprintf(file, "  motor record array   = (");

	if ( motor_record_array == NULL ) {
		fprintf(file, "NULL");
	} else {
		for ( i = 0; i < num_motors; i++ ) {
			fprintf(file, "%s", motor_record_array[i]->name);

			if ( i != num_motors-1 ) {
				fprintf(file, ", ");
			}
		}
	}
	fprintf(file, ")\n");

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_translation_motor_is_busy()";

	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	long i, num_motors;
	int busy;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;

	for ( i = 0; i < num_motors; i++ ) {
	
		child_motor_record = motor_record_array[i];

		status = mx_motor_is_busy( child_motor_record, &busy );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* If the child motor is busy, we are done. */

		if ( busy == TRUE ) {
			motor->busy = TRUE;

			return MX_SUCCESSFUL_RESULT;
		}
	}

	motor->busy = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_move_absolute()";

	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	double *motor_position_array;
	long i, num_motors;
	int simultaneous_start;
	double new_compumotor_translation_position;
	double old_compumotor_translation_position;
	double position_difference;
	double position_sum;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;
	motor_position_array
			= compumotor_translation_motor->motor_position_array;

	new_compumotor_translation_position = motor->raw_destination.analog;

	/* Check the limits on this request. */

	if ( new_compumotor_translation_position
					> motor->raw_positive_limit.analog
	  || new_compumotor_translation_position
					< motor->raw_negative_limit.analog )
	{
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Requested position %g is outside the allowed range of %g to %g",
			new_compumotor_translation_position,
			motor->raw_negative_limit.analog,
			motor->raw_positive_limit.analog );
	}

	/* Get the old positions. */

	position_sum = 0.0;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_get_position(
			child_motor_record, &motor_position_array[i] );

		if ( status.code != MXE_SUCCESS )
			return status;

		position_sum += motor_position_array[i];
	}

	old_compumotor_translation_position
				= position_sum / (double) num_motors;

	position_difference = new_compumotor_translation_position
		- old_compumotor_translation_position;

	/* Now compute the new positions. */

	for ( i = 0; i < num_motors; i++ ) {
		motor_position_array[i] += position_difference;
	}

	/* Get the MX_COMPUMOTOR_INTERFACE structure so that we can send
	 * commands directly to the controller.
	 */

	compumotor_interface_record
		= compumotor_translation_motor->compumotor_interface_record;

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The compumotor_interface_record pointer for motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			compumotor_interface_record->record_type_struct;

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_COMPUMOTOR_INTERFACE pointer for Compumotor interface '%s' is NULL.",
			compumotor_interface_record->name );
	}

	/* Are we using a simultaneous start or are we using the 
	 * linear interpolation functionality?
	 */

	if ( compumotor_translation_motor->flags
			& MXF_COMPUMOTOR_SIMULTANEOUS_START )
	{
		simultaneous_start = TRUE;
	} else {
		simultaneous_start = FALSE;
	}

	status = mxi_compumotor_multiaxis_move( compumotor_interface,
			compumotor_translation_motor->controller_number,
			num_motors, motor_record_array, motor_position_array,
			simultaneous_start );

	return status;
}


MX_EXPORT mx_status_type
mxd_compumotor_trans_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_get_position()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	double *motor_position_array;
	long i, num_motors;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	double compumotor_translation_position;
	double position_sum;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;
	motor_position_array
			= compumotor_translation_motor->motor_position_array;

	/* Get the current motor positions. */

	position_sum = 0.0;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_get_position(
			child_motor_record, &motor_position_array[i] );

		if ( status.code != MXE_SUCCESS )
			return status;

		position_sum += motor_position_array[i];
	}

	compumotor_translation_position = position_sum / (double) num_motors;

	motor->raw_position.analog = compumotor_translation_position;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_set_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"'set position' is not valid for a compumotor_translation motor." );
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_soft_abort()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	long i, num_motors;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_soft_abort( child_motor_record );
	}


	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_immediate_abort()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	long i, num_motors;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_immediate_abort( child_motor_record );
	}

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_positive_limit_hit()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	long i, num_motors;
	int limit_hit;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;

	limit_hit = FALSE;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_positive_limit_hit(
					child_motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit == TRUE ) {
			(void) mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Positive limit hit for motor '%s'",
				child_motor_record->name );

			break;             /* exit the for() loop */
		}
	}

	motor->positive_limit_hit = limit_hit;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_negative_limit_hit()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	long i, num_motors;
	int limit_hit;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;

	limit_hit = FALSE;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_negative_limit_hit(
					child_motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit == TRUE ) {
			(void) mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
				"Negative limit hit for motor '%s'",
				child_motor_record->name );

			break;             /* exit the for() loop */
		}
	}

	motor->negative_limit_hit = limit_hit;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_find_home_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
"Home searches for Compumotor_translation motor '%s' are _not_ allowed.",
		motor->record->name );
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_constant_velocity_move( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_constant_velocity_move()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	long i, num_motors;
	int direction;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	direction = motor->constant_velocity_move;

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;

	for ( i = 0; i < num_motors; i++ ) {
		child_motor_record = motor_record_array[i];

		status = mx_motor_constant_velocity_move(
					child_motor_record, direction );

		if ( status.code != MXE_SUCCESS )
			return status;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_get_parameter()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	long i, num_motors;
	double sum, max_value, double_value;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;

	switch( motor->parameter_type ) {
	case MXLV_MTR_MAXIMUM_SPEED:
		break;

	case MXLV_MTR_SPEED:
		sum = 0.0;

		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_get_speed( child_motor_record,
								&double_value );
			if ( status.code != MXE_SUCCESS )
				return status;

			sum += double_value;
		}

		motor->raw_speed = mx_divide_safely( sum, (double) num_motors );
		break;

	case MXLV_MTR_BASE_SPEED:
		sum = 0.0;

		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_get_base_speed( child_motor_record,
								&double_value );
			if ( status.code != MXE_SUCCESS )
				return status;

			sum += double_value;
		}

		motor->raw_base_speed = mx_divide_safely( sum,
							(double) num_motors );
		break;

	case MXLV_MTR_SYNCHRONOUS_MOTION_MODE:
		return mx_error( MXE_UNSUPPORTED, fname,
			"Getting the value of synchronous_motion_mode is "
			"unsupported since the pseudomotor '%s' depends on "
			"multiple motors and there is no unique answer.",
				motor->record->name );
		break;

	case MXLV_MTR_ACCELERATION_PARAMETERS:
		return mx_error( MXE_UNSUPPORTED, fname,
			"Getting the value of the acceleration parameters is "
			"unsupported since the pseudomotor '%s' depends on "
			"multiple motors and there is no unique answer.",
				motor->record->name );
		break;

	case MXLV_MTR_ACCELERATION_DISTANCE:
		max_value = 0.0;

		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_get_acceleration_distance(
					child_motor_record, &double_value );

			if ( status.code != MXE_SUCCESS )
				return status;

			if ( double_value > max_value )
				max_value = double_value;
		}

		motor->acceleration_distance = max_value;
		break;

	case MXLV_MTR_ACCELERATION_TIME:
		max_value = 0.0;

		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_get_acceleration_time(
					child_motor_record, &double_value );

			if ( status.code != MXE_SUCCESS )
				return status;

			if ( double_value > max_value )
				max_value = double_value;
		}

		motor->acceleration_time = max_value;
		break;

	case MXLV_MTR_COMPUTE_EXTENDED_SCAN_RANGE:
		return mx_motor_default_get_parameter_handler( motor );
		break;

	case MXLV_MTR_COMPUTE_PSEUDOMOTOR_POSITION:
		motor->compute_pseudomotor_position[1]
				= motor->compute_pseudomotor_position[0];
		break;

	case MXLV_MTR_COMPUTE_REAL_POSITION:
		motor->compute_real_position[1]
				= motor->compute_real_position[0];
		break;

	default:
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			motor->parameter_type );
		break;
	}

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_trans_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_trans_set_parameter()";

	MX_RECORD **motor_record_array;
	MX_RECORD *child_motor_record;
	MX_COMPUMOTOR_TRANSLATION_MOTOR *compumotor_translation_motor;
	long i, num_motors;
	double start_position, end_position, time_for_move, speed;
	mx_status_type status;

	status = mxd_compumotor_trans_get_pointers( motor,
					&compumotor_translation_motor, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	num_motors = compumotor_translation_motor->num_motors;
	motor_record_array = compumotor_translation_motor->motor_record_array;

	switch ( motor->parameter_type ) {
	case MXLV_MTR_MAXIMUM_SPEED:
		break;

	case MXLV_MTR_SPEED:
		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_set_speed( child_motor_record,
							motor->raw_speed );

			if ( status.code != MXE_SUCCESS )
				return status;
		}
		break;

	case MXLV_MTR_BASE_SPEED:
		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_set_base_speed( child_motor_record,
							motor->raw_base_speed );

			if ( status.code != MXE_SUCCESS )
				return status;
		}
		break;

	case MXLV_MTR_ACCELERATION_PARAMETERS:
		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_set_acceleration_parameters(
					child_motor_record,
					motor->acceleration_parameters );

			if ( status.code != MXE_SUCCESS )
				return status;
		}
		break;

	case MXLV_MTR_SPEED_CHOICE_PARAMETERS:
		start_position = motor->speed_choice_parameters[0];
		end_position = motor->speed_choice_parameters[1];
		time_for_move = motor->speed_choice_parameters[2];

		speed = fabs( mx_divide_safely( end_position - start_position,
							time_for_move ) );
		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_set_speed(child_motor_record, speed);

			if ( status.code != MXE_SUCCESS )
				return status;
		}
		break;

	case MXLV_MTR_SAVE_SPEED:
		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_save_speed( child_motor_record );

			if ( status.code != MXE_SUCCESS )
				return status;
		}
		break;

	case MXLV_MTR_RESTORE_SPEED:
		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_restore_speed( child_motor_record );

			if ( status.code != MXE_SUCCESS )
				return status;
		}
		break;

	case MXLV_MTR_SYNCHRONOUS_MOTION_MODE:
		for ( i = 0; i < num_motors; i++ ) {
			child_motor_record = motor_record_array[i];

			status = mx_motor_set_synchronous_motion_mode(
						child_motor_record,
						motor->synchronous_motion_mode);

			if ( status.code != MXE_SUCCESS )
				return status;
		}
		break;

	default:
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			motor->parameter_type );
		break;
	}

	return status;
}

