/*
 * Name:    d_epmcs.c 
 *
 * Purpose: MX multichannel scaler driver for the EPICS multichannel
 *          scaler support.  This currently is just for the Struck MCS.
 *
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 *
 *          Please note that the EPICS MCS record numbers scalers starting
 *          at 1, but the mcs->data_array numbers scalers starting at 0.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_meas.h"
#include "mx_mcs.h"
#include "d_epmcs.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the mcs driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_mcs_record_function_list = {
	mxd_epics_mcs_initialize_type,
	mxd_epics_mcs_create_record_structures,
	mxd_epics_mcs_finish_record_initialization,
	mxd_epics_mcs_delete_record,
	NULL,
	mxd_epics_mcs_read_parms_from_hardware,
	mxd_epics_mcs_write_parms_to_hardware,
	mxd_epics_mcs_open,
	mxd_epics_mcs_close
};

MX_MCS_FUNCTION_LIST mxd_epics_mcs_mcs_function_list = {
	mxd_epics_mcs_start,
	mxd_epics_mcs_stop,
	mxd_epics_mcs_clear,
	mxd_epics_mcs_busy,
	mxd_epics_mcs_read_all,
	mxd_epics_mcs_read_scaler,
	mxd_epics_mcs_read_measurement,
	NULL,
	mxd_epics_mcs_get_parameter,
	mxd_epics_mcs_set_parameter
};

/* EPICS mcs data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_mcs_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_MCS_STANDARD_FIELDS,
	MXD_EPICS_MCS_STANDARD_FIELDS
};

long mxd_epics_mcs_num_record_fields
		= sizeof( mxd_epics_mcs_record_field_defaults )
			/ sizeof( mxd_epics_mcs_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_mcs_rfield_def_ptr
			= &mxd_epics_mcs_record_field_defaults[0];

static mx_status_type mxd_epics_mcs_configure(
				MX_MCS *mcs, MX_EPICS_MCS *epics_mcs );

/* A private function for the use of the driver. */

static mx_status_type
mxd_epics_mcs_get_pointers( MX_MCS *mcs,
			MX_EPICS_MCS **epics_mcs,
			const char *calling_fname )
{
	const char fname[] = "mxd_epics_mcs_get_pointers()";

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCS pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( epics_mcs == (MX_EPICS_MCS **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_MCS pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( mcs->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_MCS pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*epics_mcs = (MX_EPICS_MCS *) mcs->record->record_type_struct;

	if ( *epics_mcs == (MX_EPICS_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_EPICS_MCS pointer for record '%s' is NULL.",
			mcs->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_epics_mcs_initialize_type( long record_type )
{
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	long num_record_fields;
	long maximum_num_scalers_varargs_cookie;
	long maximum_num_measurements_varargs_cookie;
	mx_status_type status;

	status = mx_mcs_initialize_type( record_type,
				&num_record_fields,
				&record_field_defaults,
				&maximum_num_scalers_varargs_cookie,
				&maximum_num_measurements_varargs_cookie );

	return status;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_mcs_create_record_structures()";

	MX_MCS *mcs;
	MX_EPICS_MCS *epics_mcs;

	/* Allocate memory for the necessary structures. */

	mcs = (MX_MCS *) malloc( sizeof(MX_MCS) );

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCS structure." );
	}

	epics_mcs = (MX_EPICS_MCS *) malloc( sizeof(MX_EPICS_MCS) );

	if ( epics_mcs == (MX_EPICS_MCS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_MCS structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = mcs;
	record->record_type_struct = epics_mcs;
	record->class_specific_function_list = &mxd_epics_mcs_mcs_function_list;

	mcs->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_finish_record_initialization( MX_RECORD *record )
{
#if 0
	const char fname[] = "mxd_epics_mcs_finish_record_initialization()";
#endif

	MX_MCS *mcs;

	mx_status_type mx_status;

	mx_status = mx_mcs_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mcs = (MX_MCS *) record->record_class_struct;

	mcs->first_scaler_number = 1;

#if 0
	MX_DEBUG(-2,("%s: record = %p", fname, record));
	MX_DEBUG(-2,("%s: record->name = '%s'", fname, record->name));
	MX_DEBUG(-2,("%s: mcs = %p", fname, mcs));
	MX_DEBUG(-2,("%s: mcs->data_array = %p", fname, mcs->data_array));
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_open( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_mcs_open()";

	MX_MCS *mcs;
	MX_EPICS_MCS *epics_mcs;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	double version_number;
	int ezca_status;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	mcs = (MX_MCS *) (record->record_class_struct);

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	/* Here we find out if this EPICS record really exists in an
	 * IOC's database by asking for the record's version number.
	 */

	strncpy( pvname, epics_mcs->channel_prefix, sizeof(pvname)-6 );
	strcat( pvname, "1.VERS" );

	MX_DEBUG( 2,("%s: about to read from '%s'.", fname, pvname));

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &version_number );

	MX_DEBUG( 2,("%s: read %g from '%s'.", fname, version_number, pvname));

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error reading EPICS PV '%s' for MX record '%s'.  EZCA status = %d (%s )",
			pvname, record->name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	epics_mcs->epics_record_version = version_number;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_epics_mcs_start( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_start()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long long_start;
	double double_start;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( epics_mcs->epics_record_version >= 5.0 ) {
		sprintf( pvname, "%sStartAll.VAL", epics_mcs->common_prefix );

		double_start = 1;

		MX_DEBUG( 2,("%s: sending %g to '%s'",
						fname, double_start, pvname));

		/* We must send this command without blocking, so we use
		 * ezcaPutOldCa().
		 */

		ezca_status = ezcaPutOldCa( pvname,
					ezcaDouble, 1, &double_start );
	} else {
		sprintf( pvname, "%s1Start.VAL", epics_mcs->channel_prefix );

		long_start = 1;

		MX_DEBUG( 2,("%s: sending %ld to '%s'",
						fname, long_start, pvname));

		/* We must send this command without blocking, so we use
		 * ezcaPutOldCa().
		 */

		ezca_status = ezcaPutOldCa( pvname, ezcaLong, 1, &long_start );
	}

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Error starting the Struck MCS '%s' (%s).  EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->channel_prefix,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_stop( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_stop()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long long_stop;
	double double_stop;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( epics_mcs->epics_record_version >= 5.0 ) {
		sprintf( pvname, "%sStopAll.VAL", epics_mcs->common_prefix );

		double_stop = 0;

		MX_DEBUG( 2,("%s: sending %g to '%s'",
					fname, double_stop, pvname));

		ezca_status = ezcaPut( pvname, ezcaDouble, 1, &double_stop );
	} else {
		sprintf( pvname, "%s1Start.VAL", epics_mcs->channel_prefix );

		long_stop = 0;

		MX_DEBUG( 2,("%s: sending %ld to '%s'",
					fname, long_stop, pvname));

		ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_stop );
	}

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error trying to stop the Struck MCS '%s' (%s).  EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->channel_prefix,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}
	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_clear( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_clear()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long long_clear;
	double double_clear;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( epics_mcs->epics_record_version >= 5.0 ) {
		sprintf( pvname, "%sEraseAll.VAL", epics_mcs->common_prefix );

		double_clear = 1;

		MX_DEBUG( 2,("%s: Sending %g to '%s'",
					fname, double_clear, pvname));

		ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_clear );
	} else {
		sprintf( pvname, "%s1.ERAS", epics_mcs->channel_prefix );

		long_clear = 1;

		MX_DEBUG( 2,("%s: Sending %ld to '%s'",
					fname, long_clear, pvname));

		ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_clear );
	}

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Error clearing the Struck MCS '%s' (%s).  EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->channel_prefix,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_busy( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_busy()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long busy;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( epics_mcs->epics_record_version >= 5.0 ) {
		sprintf( pvname, "%sAcquiring.VAL", epics_mcs->common_prefix );
	} else {
		sprintf( pvname, "%s1Start.VAL", epics_mcs->channel_prefix );
	}

	MX_DEBUG( 2,("%s: about to read from '%s'.", fname, pvname));

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &busy );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error while checking to see if the Struck MCS '%s' (%s) is still counting.  "
"EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->channel_prefix,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname,
			mx_error_message);
	}

	MX_DEBUG( 2,("%s: read %ld from '%s'.", fname, busy, pvname));

	MX_DEBUG( 2,("%s: busy = %ld", fname, busy));

	if ( busy ) {
		mcs->busy = TRUE;
	} else {
		mcs->busy = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_read_all( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_read_all()";

	long i;
	mx_status_type mx_status;

	if ( mcs == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MCS pointer passed was NULL." );
	}

	for ( i = 0; i < mcs->current_num_scalers; i++ ) {

		mcs->scaler_index = i;

		mx_status = mxd_epics_mcs_read_scaler( mcs );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_read_scaler( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_read_scaler()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long scaler_number;
	long read_cmd;
	long *data_ptr;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Please remember:  The EPICS MCS record numbers scalers starting
	 * at 1, but the mcs->data_array numbers scalers starting at 0.
	 */

	scaler_number = mcs->scaler_index + 1;

	sprintf( pvname, "%s%ld.READ",
			epics_mcs->channel_prefix, scaler_number );

	read_cmd = 1;

	MX_DEBUG( 2,("%s: sending %ld to '%s'", fname, read_cmd, pvname));

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &read_cmd );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error reading values for Struck MCS channel %ld.  EZCA status = %d (%s )",
			scaler_number, ezca_status, ezca_error_message );

		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	sprintf( pvname, "%s%ld.VAL",
			epics_mcs->channel_prefix, scaler_number );

#if 0
	MX_DEBUG(-2,("%s: mcs->record = %p", fname, mcs->record));
	MX_DEBUG(-2,("%s: mcs->record->name = '%s'", fname, mcs->record->name));
	MX_DEBUG(-2,("%s: mcs = %p", fname, mcs));
	MX_DEBUG(-2,("%s: mcs->data_array = %p", fname, mcs->data_array));
	MX_DEBUG(-2,("%s: mcs->scaler_index = %ld", fname, mcs->scaler_index));
#endif

	data_ptr = mcs->data_array[ mcs->scaler_index ];

#if 0
	MX_DEBUG(-2,("%s: data_ptr = %p", fname, data_ptr));

	MX_DEBUG(-2,("%s: mcs->num_measurements = %ld",
			fname, mcs->num_measurements));
#endif

	MX_DEBUG( 2,("%s: about to read from '%s'.", fname, pvname));

	ezca_status = ezcaGet( pvname, ezcaLong, 
				mcs->current_num_measurements, data_ptr );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error transferring data for Struck MCS channel %ld.  EZCA status = %d (%s )",
			scaler_number, ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	MX_DEBUG( 2,("%s: data received for '%s'.", fname, pvname));

#if 0
	{
		long i;

		for ( i = 0; i < mcs->num_measurements; i++ ) {
			fprintf(stderr,"%ld ", data_ptr[i]);
		}
		fprintf(stderr,"\n");
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_read_measurement( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_read_measurement()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"This function is not supported for EPICS multichannel scalers." );
}

MX_EXPORT mx_status_type
mxd_epics_mcs_get_parameter( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_get_parameter()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_epics_mcs_set_parameter( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_set_parameter()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	double dwell_time, preset_live_time;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for MCS '%s', parameter type '%s' (%d)",
		fname, mcs->record->name,
		mx_get_field_label_string( mcs->record, mcs->parameter_type ),
		mcs->parameter_type));

	if ( mcs->parameter_type == MXLV_MCS_MODE ) {

		if ( mcs->mode != MXM_PRESET_TIME ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal MCS mode %d selected.  Only preset time mode is "
		"allowed for an EPICS MCS.", mcs->mode );
		}

		mx_status = mxd_epics_mcs_configure( mcs, epics_mcs );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

	} else if ( mcs->parameter_type == MXLV_MCS_MEASUREMENT_TIME ) {

		if ( epics_mcs->epics_record_version >= 5.0 ) {
			sprintf( pvname, "%sDwell.VAL",
						epics_mcs->common_prefix );
		} else {
			sprintf( pvname, "%s1.DWEL",
						epics_mcs->channel_prefix );
		}

		dwell_time = mcs->measurement_time;

		MX_DEBUG( 2,("%s: sending %g to '%s'",
					fname, dwell_time, pvname));

		ezca_status = ezcaPut( pvname, ezcaDouble, 1, &dwell_time );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error setting EPICS MCS '%s' (%s) dwell time.  EZCA status = %d (%s )",
				mcs->record->name,
				epics_mcs->channel_prefix,
				ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR,
					fname, mx_error_message);
		}

	} else if ( mcs->parameter_type == MXLV_MCS_CURRENT_NUM_MEASUREMENTS ){

		if ( epics_mcs->epics_record_version >= 5.0 ) {
			sprintf( pvname, "%sNuseAll.VAL",
						epics_mcs->common_prefix );

			MX_DEBUG( 2,("%s: sending %ld to '%s'",
					fname, mcs->current_num_measurements,
					pvname));

			ezca_status = ezcaPut( pvname, ezcaLong,
					1, &(mcs->current_num_measurements) );

			if ( ezca_status != EZCA_OK ) {
				ezcaGetErrorString( NULL, &ezca_error_message );
				sprintf( mx_error_message,
		"Error setting EPICS MCS '%s' (%s) number of measurements.  "
		"EZCA status = %d (%s )", mcs->record->name,
					epics_mcs->channel_prefix,
					ezca_status, ezca_error_message );
				ezcaFree( (void *) ezca_error_message );

				return mx_error(MXE_DEVICE_IO_ERROR,
					fname, mx_error_message);
			}
		} else {
			sprintf( pvname, "%s1.PLTM",
						epics_mcs->channel_prefix );

			preset_live_time = mcs->measurement_time
				* (double) mcs->current_num_measurements;

			MX_DEBUG( 2,("%s: sending %g to '%s'",
					fname, preset_live_time, pvname));

			ezca_status = ezcaPut( pvname, ezcaDouble,
						1, &preset_live_time );

			if ( ezca_status != EZCA_OK ) {
				ezcaGetErrorString( NULL, &ezca_error_message );
				sprintf( mx_error_message,
"Error setting EPICS MCS '%s' (%s) preset live time.  EZCA status = %d (%s )",
					mcs->record->name,
					epics_mcs->channel_prefix,
					ezca_status, ezca_error_message );
				ezcaFree( (void *) ezca_error_message );

				return mx_error(MXE_DEVICE_IO_ERROR,
					fname, mx_error_message);
			}
		}

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			mcs->parameter_type );
	}
	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

static mx_status_type
mxd_epics_mcs_configure( MX_MCS *mcs, MX_EPICS_MCS *epics_mcs )
{
#if 0
	const char fname[] = "mxd_epics_mcs_configure()";

	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long long_value;
	double double_value;

	MX_DEBUG( 2,("%s invoked.", fname));

	/* Put the Struck EPICS record into MCS mode. */

	if ( epics_mcs->epics_record_version < 5.0 ) {
		sprintf( pvname, "%s1.MODE", epics_mcs->channel_prefix );

		long_value = 1;

		MX_DEBUG( 2,("%s: sending %ld to '%s'",
					fname, long_value, pvname));

		ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_value );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error putting EPICS MCS '%s' (%s) into MCS mode.  EZCA status = %d (%s )",
				mcs->record->name,
				epics_mcs->channel_prefix,
				ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR,
						fname, mx_error_message);
		}
	}

	/* Set the channel advance source for the Struck MCS. */

	if ( epics_mcs->epics_record_version >= 5.0 ) {
		sprintf( pvname, "%sChannelAdvance.VAL",
					epics_mcs->common_prefix );
	} else {
		sprintf( pvname, "%s1.CHAS", epics_mcs->channel_prefix );
	}

	long_value = epics_mcs->use_external_channel_advance;

	MX_DEBUG( 2,("%s: sending %ld to '%s'", fname, long_value, pvname));

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error setting the channel advance source for the EPICS MCS '%s' (%s).  "
"EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->channel_prefix,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Set the external prescale for the Struck MCS. */

	if ( epics_mcs->epics_record_version >= 5.0 ) {
		sprintf( pvname, "%sPrescale.VAL", epics_mcs->common_prefix );

		double_value = 0;

		MX_DEBUG( 2,("%s: sending %g to '%s'",
					fname, double_value, pvname));

		ezca_status = ezcaPut( pvname, ezcaDouble, 1, &double_value );
	} else {
		sprintf( pvname, "%s1.PSCL", epics_mcs->channel_prefix );

		long_value = 0;

		MX_DEBUG( 2,("%s: sending %ld to '%s'",
					fname, long_value, pvname));

		ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_value );
	}

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error setting the prescale for the EPICS MCS '%s' (%s).  "
"EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->channel_prefix,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}
#endif
	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

