/*
 * Name:    d_nmotor.c 
 *
 * Purpose: MX motor driver for motors controlled via an MX network server.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_motor.h"
#include "d_nmotor.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_network_motor_record_function_list = {
	mxd_network_motor_initialize_type,
	mxd_network_motor_create_record_structures,
	mxd_network_motor_finish_record_initialization,
	mxd_network_motor_delete_record,
	mxd_network_motor_print_motor_structure,
	mxd_network_motor_read_parms_from_hardware,
	mxd_network_motor_write_parms_to_hardware,
	mxd_network_motor_open,
	mxd_network_motor_close,
	NULL,
	mxd_network_motor_resynchronize
};

MX_MOTOR_FUNCTION_LIST mxd_network_motor_motor_function_list = {
	mxd_network_motor_motor_is_busy,
	mxd_network_motor_move_absolute,
	mxd_network_motor_get_position,
	mxd_network_motor_set_position,
	mxd_network_motor_soft_abort,
	mxd_network_motor_immediate_abort,
	mxd_network_motor_positive_limit_hit,
	mxd_network_motor_negative_limit_hit,
	mxd_network_motor_find_home_position,
	mxd_network_motor_constant_velocity_move,
	mxd_network_motor_get_parameter,
	mxd_network_motor_set_parameter
};

/* Soft motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_network_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_NETWORK_MOTOR_STANDARD_FIELDS
};

long mxd_network_motor_num_record_fields
		= sizeof( mxd_network_motor_record_field_defaults )
			/ sizeof( mxd_network_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_motor_rfield_def_ptr
			= &mxd_network_motor_record_field_defaults[0];

/*=======================================================================*/

/* This function is a utility function to consolidate all of the
 * pointer mangling that often has to happen at the beginning of an 
 * mxd_network_motor_...  function.  The parameter 'calling_fname'
 * is passed so that error messages will appear with the name of
 * the calling function.
 */

MX_EXPORT mx_status_type
mxd_network_motor_get_pointers( MX_MOTOR *motor,
			MX_NETWORK_MOTOR **network_motor,
			const char *calling_fname )
{
	const char fname[] = "mxd_network_motor_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The motor pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( motor->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_RECORD pointer for motor pointer passed by '%s' is NULL.",
			calling_fname );
	}

	if ( motor->record->mx_type != MXT_MTR_NETWORK ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The motor '%s' passed by '%s' is not a network motor.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			motor->record->name, calling_fname,
			motor->record->mx_superclass,
			motor->record->mx_class,
			motor->record->mx_type );
	}

	if ( network_motor != (MX_NETWORK_MOTOR **) NULL ) {

		*network_motor = (MX_NETWORK_MOTOR *)
				(motor->record->record_type_struct);

		if ( *network_motor == (MX_NETWORK_MOTOR *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_NETWORK_MOTOR pointer for motor record '%s' passed by '%s' is NULL",
				motor->record->name, calling_fname );
		}
	}

	if ( (*network_motor)->server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX server record pointer for MX network motor '%s' is NULL.",
			motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mxd_network_motor_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_network_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_NETWORK_MOTOR *network_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	network_motor = (MX_NETWORK_MOTOR *) malloc( sizeof(MX_NETWORK_MOTOR) );

	if ( network_motor == (MX_NETWORK_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_NETWORK_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = network_motor;
	record->class_specific_function_list
				= &mxd_network_motor_motor_function_list;

	motor->record = record;

	/* A network motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_motor_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_network_motor_finish_record_initialization()";

	MX_MOTOR *motor;
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	motor = (MX_MOTOR *) record->record_class_struct;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	motor->flags |= MXF_MTR_IS_REMOTE_MOTOR;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_motor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_motor_print_motor_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_network_motor_print_motor_structure()";

	MX_MOTOR *motor;
	MX_NETWORK_MOTOR *network_motor;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE,
			"MX_MOTOR pointer for record '%s' is NULL.",
			record->name);
	}

	network_motor = (MX_NETWORK_MOTOR *) (record->record_type_struct);

	if ( network_motor == (MX_NETWORK_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE,
			"MX_NETWORK_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type     = NETWORK_MOTOR.\n\n");

	fprintf(file, "  name           = %s\n", record->name);
	fprintf(file, "  server         = %s\n",
					network_motor->server_record->name );
	fprintf(file, "  remote record  = %s\n",
					network_motor->remote_record_name );

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS )
		return status;

	fprintf(file, "  position       = %g %s (%g)\n",
		motor->position, motor->units, motor->raw_position.analog );

	fprintf(file, "  scale          = %g %s per step.\n",
		motor->scale, motor->units );

	fprintf(file, "  offset         = %g %s.\n",
		motor->offset, motor->units );

	fprintf(file, "  backlash       = %g %s (%g)\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog );

	fprintf(file, "  negative_limit = %g %s (%g)\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog );

	fprintf(file, "  positive_limit = %g %s (%g)\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.analog );

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband       = %g %s (%g)\n\n",
		deadband, motor->units,
		motor->deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_motor_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_motor_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_motor_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_motor_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_motor_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxd_network_motor_resynchronize()";

	MX_MOTOR *motor;
	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int resynchronize;
	mx_status_type status;

	motor = (MX_MOTOR *) record->record_class_struct;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS ) {
		motor->busy = FALSE;

		return status;
	}

	sprintf(rfname, "%s.resynchronize", network_motor->remote_record_name);

	resynchronize = 1;

	status = mx_put( network_motor->server_record,
				rfname, MXFT_INT, &resynchronize );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_motor_is_busy()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int busy;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS ) {
		motor->busy = FALSE;

		return status;
	}

	sprintf( rfname, "%s.busy", network_motor->remote_record_name );

	status = mx_get( network_motor->server_record,
				rfname, MXFT_INT, &busy );

	motor->busy = busy;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_move_absolute()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	double new_destination;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	new_destination = motor->raw_destination.analog;

	sprintf( rfname, "%s.destination", network_motor->remote_record_name );

	status = mx_put( network_motor->server_record,
				rfname, MXFT_DOUBLE, &new_destination );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_get_position()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	double position;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.position", network_motor->remote_record_name );

	status = mx_get( network_motor->server_record,
				rfname, MXFT_DOUBLE, &position );

	motor->raw_position.analog = position;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_set_position()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	double new_set_position;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	new_set_position = motor->raw_set_position.analog;

	sprintf( rfname, "%s.set_position", network_motor->remote_record_name );

	status = mx_put( network_motor->server_record,
				rfname, MXFT_DOUBLE, &new_set_position );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_soft_abort()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int abort_flag;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	abort_flag = 1;

	sprintf( rfname, "%s.soft_abort", network_motor->remote_record_name );

	status = mx_put( network_motor->server_record,
				rfname, MXFT_INT, &abort_flag );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_immediate_abort()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int abort_flag;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	abort_flag = 1;

	sprintf( rfname, "%s.immediate_abort",
				network_motor->remote_record_name );

	status = mx_put( network_motor->server_record,
				rfname, MXFT_INT, &abort_flag );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_positive_limit_hit()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int limit_hit;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.positive_limit_hit",
				network_motor->remote_record_name );

	status = mx_get( network_motor->server_record,
				rfname, MXFT_INT, &limit_hit );

	motor->positive_limit_hit = limit_hit;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_negative_limit_hit()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int limit_hit;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.negative_limit_hit",
				network_motor->remote_record_name );

	status = mx_get( network_motor->server_record,
				rfname, MXFT_INT, &limit_hit );

	motor->negative_limit_hit = limit_hit;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_find_home_position()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int home_search;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	home_search = motor->home_search;

	sprintf( rfname, "%s.home_search", network_motor->remote_record_name );

	status = mx_put( network_motor->server_record,
				rfname, MXFT_INT, &home_search );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_constant_velocity_move( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_constant_velocity_move()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	int constant_velocity_move;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	constant_velocity_move = motor->constant_velocity_move;

	sprintf( rfname, "%s.constant_velocity_move",
					network_motor->remote_record_name );

	status = mx_put( network_motor->server_record,
				rfname, MXFT_INT, &constant_velocity_move );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_get_parameter()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension_array[1];
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( rfname, "%s.speed",
				network_motor->remote_record_name );

		status = mx_get( network_motor->server_record,
				rfname, MXFT_DOUBLE, &(motor->raw_speed) );
		break;

	case MXLV_MTR_BASE_SPEED:
		sprintf( rfname, "%s.base_speed",
				network_motor->remote_record_name );

		status = mx_get( network_motor->server_record,
				rfname, MXFT_DOUBLE, &(motor->raw_base_speed) );
		break;

	case MXLV_MTR_MAXIMUM_SPEED:
		sprintf( rfname, "%s.maximum_speed",
				network_motor->remote_record_name );

		status = mx_get( network_motor->server_record,
				rfname, MXFT_DOUBLE,
				&(motor->raw_maximum_speed) );
		break;

	case MXLV_MTR_SYNCHRONOUS_MOTION_MODE:
		sprintf( rfname, "%s.synchronous_motion_mode",
				network_motor->remote_record_name );

		status = mx_get( network_motor->server_record,
				rfname, MXFT_INT,
				&(motor->synchronous_motion_mode) );
		break;

	case MXLV_MTR_ACCELERATION_PARAMETERS:
		sprintf( rfname, "%s.acceleration_parameters",
				network_motor->remote_record_name );

		dimension_array[0] = MX_MOTOR_NUM_ACCELERATION_PARAMS;

		status = mx_get_array( network_motor->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&(motor->acceleration_parameters) );
		break;

	case MXLV_MTR_ACCELERATION_DISTANCE:
		sprintf( rfname, "%s.acceleration_distance",
				network_motor->remote_record_name );

		status = mx_get( network_motor->server_record,
				rfname, MXFT_DOUBLE,
				&(motor->acceleration_distance) );
		break;

	case MXLV_MTR_ACCELERATION_TIME:
		sprintf( rfname, "%s.acceleration_time",
				network_motor->remote_record_name );

		status = mx_get( network_motor->server_record,
				rfname, MXFT_DOUBLE,
				&(motor->acceleration_time) );
		break;

	case MXLV_MTR_COMPUTE_EXTENDED_SCAN_RANGE:
		sprintf( rfname, "%s.compute_extended_scan_range",
				network_motor->remote_record_name );

		dimension_array[0] = MX_MOTOR_NUM_SCAN_RANGE_PARAMS;

		/* Send the original scan range limits. */

		status = mx_put_array( network_motor->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&(motor->compute_extended_scan_range) );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* Then, get the extended scan range limits. */

		status = mx_get_array( network_motor->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&(motor->compute_extended_scan_range) );
		break;

	case MXLV_MTR_COMPUTE_PSEUDOMOTOR_POSITION:
		sprintf( rfname, "%s.compute_pseudomotor_position",
				network_motor->remote_record_name );

		dimension_array[0] = MX_MOTOR_NUM_POSITION_PARAMS;

		/* Send the real motor position. */

		status = mx_put_array( network_motor->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&(motor->compute_pseudomotor_position) );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* Get the pseudomotor position. */

		status = mx_get_array( network_motor->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&(motor->compute_pseudomotor_position) );
		break;

	case MXLV_MTR_COMPUTE_REAL_POSITION:
		sprintf( rfname, "%s.compute_real_position",
				network_motor->remote_record_name );

		dimension_array[0] = MX_MOTOR_NUM_POSITION_PARAMS;

		/* Send the pseudomotor position. */

		status = mx_put_array( network_motor->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&(motor->compute_real_position) );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* Get the real motor position. */

		status = mx_get_array( network_motor->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&(motor->compute_real_position) );
		break;

	default:
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			motor->parameter_type );
		break;
	}

	return status;
}

MX_EXPORT mx_status_type
mxd_network_motor_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_network_motor_set_parameter()";

	MX_NETWORK_MOTOR *network_motor;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH + 1 ];
	long dimension_array[1];
	int int_value;
	mx_status_type status;

	status = mxd_network_motor_get_pointers(motor, &network_motor, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 2,("%s invoked for motor '%s' for parameter type '%s' (%d).",
		fname, motor->record->name,
		mx_get_field_label_string( motor->record,
			motor->parameter_type ),
		motor->parameter_type));

	switch ( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( rfname, "%s.speed",
				network_motor->remote_record_name );

		status = mx_put( network_motor->server_record,
				rfname, MXFT_DOUBLE, &(motor->raw_speed) );
		break;

	case MXLV_MTR_BASE_SPEED:
		sprintf( rfname, "%s.base_speed",
				network_motor->remote_record_name );

		status = mx_put( network_motor->server_record,
				rfname, MXFT_DOUBLE, &(motor->raw_base_speed) );
		break;

	case MXLV_MTR_MAXIMUM_SPEED:
		sprintf( rfname, "%s.maximum_speed",
				network_motor->remote_record_name );

		status = mx_put( network_motor->server_record,
				rfname, MXFT_DOUBLE,
				&(motor->raw_maximum_speed) );
		break;

	case MXLV_MTR_ACCELERATION_PARAMETERS:
		sprintf( rfname, "%s.acceleration_parameters",
				network_motor->remote_record_name );

		dimension_array[0] = MX_MOTOR_NUM_ACCELERATION_PARAMS;

		status = mx_put_array( network_motor->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&(motor->acceleration_parameters) );
		break;

	case MXLV_MTR_SPEED_CHOICE_PARAMETERS:
		sprintf( rfname, "%s.speed_choice_parameters",
				network_motor->remote_record_name );

		dimension_array[0] = MX_MOTOR_NUM_SPEED_CHOICE_PARAMS;

		status = mx_put_array( network_motor->server_record,
				rfname, MXFT_DOUBLE, 1, dimension_array,
				&(motor->speed_choice_parameters) );
		break;

	case MXLV_MTR_SAVE_SPEED:
		sprintf( rfname, "%s.save_speed",
				network_motor->remote_record_name );

		int_value = 1;

		status = mx_put( network_motor->server_record,
				rfname, MXFT_INT, &int_value );
		break;

	case MXLV_MTR_RESTORE_SPEED:
		sprintf( rfname, "%s.restore_speed",
				network_motor->remote_record_name );

		int_value = 1;

		status = mx_put( network_motor->server_record,
				rfname, MXFT_INT, &int_value );
		break;

	case MXLV_MTR_SYNCHRONOUS_MOTION_MODE:
		sprintf( rfname, "%s.synchronous_motion_mode",
				network_motor->remote_record_name );

		status = mx_put( network_motor->server_record,
				rfname, MXFT_INT,
				&(motor->synchronous_motion_mode) );
		break;

	default:
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			motor->parameter_type );
		break;
	}

	return status;
}

