/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_s10id.h 
 *
 * Purpose: Header file for MX motor driver to control the MRCAT Sector 10-ID
 *          monochromator at the Advanced Photon Source.  This is a pseudomotor
 *          that pretends to be the monochromator 'theta' axis.  The value
 *          written to this pseudo motor is used to control the real 'theta'
 *          axis and any other MX devices that are tied into this pseudomotor.
 */

#ifndef __D_S10ID_H__
#define __D_S10ID_H__

#include "mx_motor.h"

/* ===== MX Sector 10-ID motor data structures ===== */

typedef struct {
	MX_RECORD *real_theta_motor_record;
	MX_RECORD *omega_motor_record;
	MX_RECORD *omega_function_type_record;
	MX_RECORD *omega_parameters_record;
	MX_RECORD *gap_energy_motor_record;
	MX_RECORD *gap_linked_record;
	MX_RECORD *gap_offset_record;
	MX_RECORD *gap_harmonic_record;
	MX_RECORD *d_spacing_record;
	double angle_scale;  /* To translate betw. radians and native units. */
} MX_S10ID_MOTOR;

#define MXF_S10ID_NUM_REAL_MOTOR_RECORDS 3

/* Omega function types. */

#define MXF_S10ID_NO_DEPENDENCE		0
#define MXF_S10ID_CUBIC_POLYNOMIAL_FIT	1

/* Define all of the interface functions. */

MX_API mx_status_type mxd_s10id_motor_initialize_type( long type );
MX_API mx_status_type mxd_s10id_motor_create_record_structures(
					MX_RECORD *record );
MX_API mx_status_type mxd_s10id_motor_finish_record_initialization(
					MX_RECORD *record );
MX_API mx_status_type mxd_s10id_motor_delete_record( MX_RECORD *record );
MX_API mx_status_type mxd_s10id_motor_print_motor_structure(
					FILE *file, MX_RECORD *record );
MX_API mx_status_type mxd_s10id_motor_read_parms_from_hardware(
						MX_RECORD *record );
MX_API mx_status_type mxd_s10id_motor_write_parms_to_hardware(
						MX_RECORD *record );
MX_API mx_status_type mxd_s10id_motor_open( MX_RECORD *record );
MX_API mx_status_type mxd_s10id_motor_close( MX_RECORD *record );
MX_API mx_status_type mxd_s10id_motor_motor_is_busy( MX_MOTOR *motor );
MX_API mx_status_type mxd_s10id_motor_move_absolute( MX_MOTOR *motor );
MX_API mx_status_type mxd_s10id_motor_get_position( MX_MOTOR *motor );
MX_API mx_status_type mxd_s10id_motor_set_position( MX_MOTOR *motor );
MX_API mx_status_type mxd_s10id_motor_soft_abort( MX_MOTOR *motor );
MX_API mx_status_type mxd_s10id_motor_immediate_abort( MX_MOTOR *motor );
MX_API mx_status_type mxd_s10id_motor_positive_limit_hit( MX_MOTOR *motor );
MX_API mx_status_type mxd_s10id_motor_negative_limit_hit( MX_MOTOR *motor );
MX_API mx_status_type mxd_s10id_motor_find_home_position( MX_MOTOR *motor );

extern MX_RECORD_FUNCTION_LIST mxd_s10id_motor_record_function_list;
extern MX_MOTOR_FUNCTION_LIST mxd_s10id_motor_motor_function_list;

extern long mxd_s10id_motor_num_record_fields;
extern MX_RECORD_FIELD_DEFAULTS *mxd_s10id_motor_rfield_def_ptr;

#define MXD_S10ID_MOTOR_STANDARD_FIELDS \
  {-1, -1, "real_theta_motor_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, \
		offsetof(MX_S10ID_MOTOR, real_theta_motor_record), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY) }, \
  \
  {-1, -1, "omega_motor_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, offsetof(MX_S10ID_MOTOR, omega_motor_record), \
	{0}, NULL, MXFF_IN_DESCRIPTION }, \
  \
  {-1, -1, "omega_function_type_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, \
		offsetof(MX_S10ID_MOTOR, omega_function_type_record), \
	{0}, NULL, MXFF_IN_DESCRIPTION }, \
  \
  {-1, -1, "omega_parameters_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, \
		offsetof(MX_S10ID_MOTOR, omega_parameters_record), \
	{0}, NULL, MXFF_IN_DESCRIPTION }, \
  \
  {-1, -1, "gap_energy_motor_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, \
		offsetof(MX_S10ID_MOTOR, gap_energy_motor_record), \
	{0}, NULL, MXFF_IN_DESCRIPTION }, \
  \
  {-1, -1, "gap_linked_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, offsetof(MX_S10ID_MOTOR, gap_linked_record), \
	{0}, NULL, MXFF_IN_DESCRIPTION }, \
  \
  {-1, -1, "gap_offset_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, offsetof(MX_S10ID_MOTOR, gap_offset_record), \
	{0}, NULL, MXFF_IN_DESCRIPTION }, \
  \
  {-1, -1, "gap_harmonic_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, offsetof(MX_S10ID_MOTOR, gap_harmonic_record), \
	{0}, NULL, MXFF_IN_DESCRIPTION }, \
  \
  {-1, -1, "d_spacing_record", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, offsetof(MX_S10ID_MOTOR, d_spacing_record), \
	{0}, NULL, MXFF_IN_DESCRIPTION }, \
  \
  {-1, -1, "angle_scale", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, offsetof(MX_S10ID_MOTOR, angle_scale), \
	{0}, NULL, MXFF_IN_DESCRIPTION }

#endif /* __D_S10ID_H__ */
