/*
 * Name:    i_xia_xerxes.c
 *
 * Purpose: MX driver for the X-Ray Instrumentation Associates 
 *          XerXes library used by the DXP-2X multichannel analyzer.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_XIA_XERXES

#include <stdlib.h>
#include <errno.h>

#include <xerxes_structures.h>
#include <xerxes_errors.h>
#include <xerxes.h>

#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_types.h"
#include "mx_alloc.h"
#include "mx_mca.h"
#include "i_xia_xerxes.h"
#include "d_xia_dxp.h"

MX_RECORD_FUNCTION_LIST mxi_xia_xerxes_record_function_list = {
	mxi_xia_xerxes_initialize_type,
	mxi_xia_xerxes_create_record_structures,
	mxi_xia_xerxes_finish_record_initialization,
	mxi_xia_xerxes_delete_record,
	NULL,
	mxi_xia_xerxes_read_parms_from_hardware,
	mxi_xia_xerxes_write_parms_to_hardware,
	mxi_xia_xerxes_open,
	mxi_xia_xerxes_close
};

MX_RECORD_FIELD_DEFAULTS mxi_xia_xerxes_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_XIA_XERXES_STANDARD_FIELDS
};

long mxi_xia_xerxes_num_record_fields
		= sizeof( mxi_xia_xerxes_record_field_defaults )
			/ sizeof( mxi_xia_xerxes_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_xia_xerxes_rfield_def_ptr
			= &mxi_xia_xerxes_record_field_defaults[0];

#define MXI_XIA_XERXES_DEBUG			TRUE

/* A private function for the use of the driver. */

static mx_status_type
mxi_xia_xerxes_get_pointers( MX_MCA *mca,
			MX_XIA_DXP_MCA **xia_dxp_mca,
			MX_XIA_XERXES **xia_xerxes,
			const char *calling_fname )
{
	const char fname[] = "mxi_xia_xerxes_get_pointers()";

	MX_RECORD *xia_dxp_record;

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_dxp_mca == (MX_XIA_DXP_MCA **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_DXP_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( xia_xerxes == (MX_XIA_XERXES **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_XIA_XERXES pointer passed by '%s' was NULL.",
			calling_fname );
	}
	if ( mca->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX_RECORD pointer for the MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*xia_dxp_mca = (MX_XIA_DXP_MCA *) mca->record->record_type_struct;

	xia_dxp_record = (*xia_dxp_mca)->xia_dxp_record;

	if ( xia_dxp_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The xia_dxp_record pointer for MCA record '%s' is NULL.",
			mca->record->name );
	}

	*xia_xerxes = (MX_XIA_XERXES *) xia_dxp_record->record_type_struct;

	if ( (*xia_xerxes) == (MX_XIA_XERXES *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_XERXES pointer for XIA DXP record '%s' is NULL.",
			xia_dxp_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==========================*/

MX_EXPORT mx_status_type
mxi_xia_xerxes_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_xerxes_create_record_structures()";

	MX_XIA_XERXES *xia_xerxes;

	/* Allocate memory for the necessary structures. */

	xia_xerxes = (MX_XIA_XERXES *) malloc( sizeof(MX_XIA_XERXES) );

	if ( xia_xerxes == (MX_XIA_XERXES *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Can't allocate memory for MX_XIA_XERXES structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = NULL;;
	record->record_type_struct = xia_xerxes;

	record->class_specific_function_list = NULL;

	xia_xerxes->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_xerxes_finish_record_initialization()";

	MX_XIA_XERXES *xia_xerxes;

	xia_xerxes = (MX_XIA_XERXES *) record->record_type_struct;

	if ( xia_xerxes == (MX_XIA_XERXES *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_XIA_XERXES pointer for record '%s' is NULL.",
			record->name );
	}

	/* We cannot allocate the mca_record_array structures until
	 * mxi_xia_xerxes_open() is invoked, so for now we initialize
	 * these data structures to all zeros.
	 */

	xia_xerxes->num_mcas = 0;
	xia_xerxes->mca_record_array = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_open( MX_RECORD *record )
{
	const char fname[] = "mxi_xia_xerxes_open()";

	MX_XIA_XERXES *xia_xerxes;
	int xia_status;
	int config_file_lun, config_file_mode;
	int i, num_mcas;
	Xia_Util_Functions md_util_struct;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	xia_xerxes = (MX_XIA_XERXES *) (record->record_type_struct);

	if ( xia_xerxes == (MX_XIA_XERXES *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_XIA_XERXES pointer for record '%s' is NULL.", record->name);
	}

	/* Get pointers to the XerXes MD utility functions we will use. */

	xia_status = dxp_md_init_util( &md_util_struct, "Unused");

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to get pointers to the XerXes MD utility functions.  "
		"Error code = %d, error status = '%s'",
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	xia_xerxes->dxp_md_error = md_util_struct.dxp_md_error;
	xia_xerxes->dxp_md_alloc = md_util_struct.dxp_md_alloc;
	xia_xerxes->dxp_md_free  = md_util_struct.dxp_md_free;
	xia_xerxes->dxp_md_puts  = md_util_struct.dxp_md_puts;
	xia_xerxes->dxp_md_wait  = md_util_struct.dxp_md_wait;

	/* Initialize XerXes. */

	xia_status = dxp_init_library();

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to initialize XerXes.  "
		"Error code = %d, error status = '%s'",
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Read the modules configuration file. */

	xia_status = dxp_read_config( xia_xerxes->modules_filename );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to read the modules configuration from "
		"the XerXes configuration file '%s'.  "
		"Error code = %d, error status = '%s'",
			xia_xerxes->modules_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Open the detector configuration file. */

	config_file_mode = 2;

	xia_status = dxp_open_file( &config_file_lun,
					xia_xerxes->config_filename,
					&config_file_mode );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_FILE_IO_ERROR, fname,
		"Unable to open the XerXes detector configuration file '%s'.  "
		"Error code = %d, error status = '%s'",
			xia_xerxes->config_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Restore the detector configuration. */

	xia_status = dxp_restore_config( &config_file_lun );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to restore the detector configuration from "
		"the XerXes configuration file '%s'.  "
		"Error code = %d, error status = '%s'",
			xia_xerxes->config_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Close the detector configuration file. */

	xia_status = dxp_close_file( &config_file_lun );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_FILE_IO_ERROR, fname,
		"Unable to close the XerXes detector configuration file '%s'.  "
		"Error code = %d, error status = '%s'",
			xia_xerxes->config_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Send a stop run command to make sure that all DXP channels
	 * are stopped.
	 */

	xia_status = dxp_stop_run();

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_FILE_IO_ERROR, fname,
		"Unable to stop runs on all DXP channels.  "
		"Error code = %d, error status = '%s'",
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* Find out how many detector channels (MCAs) there are. */

	xia_status = dxp_ndxpchan( &num_mcas );

	MX_DEBUG(-2,("%s: num_mcas = %d", fname, num_mcas));

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Unable to find out how many detector channels are in "
		"the current configuration '%s'.  "
		"Error code = %d, error status = '%s'",
			xia_xerxes->config_filename,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	xia_xerxes->mca_record_array = ( MX_RECORD ** ) malloc(
					num_mcas * sizeof( MX_RECORD * ) );

	if ( xia_xerxes->mca_record_array == ( MX_RECORD ** ) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %d element array "
		"of MX_RECORD pointers for the mca_record_array data "
		"structure of record '%s'.", num_mcas, record->name );
	}

	for ( i = 0; i < num_mcas; i++ ) {
		xia_xerxes->mca_record_array[i] = NULL;
	}

	xia_xerxes->num_mcas = num_mcas;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxi_xia_xerxes_is_busy( MX_MCA *mca,
			int *busy_flag,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_read_status()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	int xia_status, active;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_status = dxp_isrunning( &(xia_dxp_mca->system_channel), &active );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot get DXP running status for MCA '%s'.  "
		"Error code = %d, error status = '%s'",
			mca->record->name, xia_status,
			mxi_xia_xerxes_strerror( xia_status ) );
	}

	/* We are only interested in the two lowest order bits.
	 *
	 * Bit 0 tells you whether or not the board things a run
	 * is active.  Bit 1 tells you whether or not XerXes
	 * thinks a run is active.
	 */

	active &= 0x3;

	MX_DEBUG(-2,("%s: active = %#x", fname, active));

	switch( active ) {
	case 0x0:
		*busy_flag = FALSE;
		break;
	case 0x1:
		(void) mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Inconsistent active state flags %#x: "
		"The DXP board for MCA '%s' says that it is active, "
		"but XerXes says that it is not.",
			active, mca->record->name );
		*busy_flag = TRUE;
		break;
	case 0x2:
		(void) mx_error( MXE_INTERFACE_ACTION_FAILED, fname,
		"Inconsistent active state flags %#x: "
		"XerXes says that MCA '%s' is active, but the DXP board "
		"says that it is not.",
			active, mca->record->name );
		*busy_flag = FALSE;
		break;
	case 0x3:
		*busy_flag = TRUE;
		break;
	default:
		return mx_error( MXE_FUNCTION_FAILED, fname,
		"It should not be possible to get to this message.  "
		"If you do, then something peculiar has happened "
		"for MCA '%s' and this should be reported to the programmer.",
			mca->record->name );
		break;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_read_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type *value_ptr,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_read_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned short short_value;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Record '%s', channel %d, parameter '%s'.",
			fname, mca->record->name,
			xia_dxp_mca->system_channel, parameter_name));
	}

	xia_status = dxp_get_one_dspsymbol( &(xia_dxp_mca->system_channel),
						parameter_name, &short_value );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot read DXP parameter '%s' for system channel %d "
		"of XIA DXP interface '%s'.  "
		"Error code = %d, error status = '%s'",
			parameter_name, xia_dxp_mca->system_channel,
			mca->record->name, xia_status,
			mxi_xia_xerxes_strerror( xia_status ) );
	}

	*value_ptr = (mx_uint32_type) short_value;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: value = %lu", fname, *value_ptr));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_write_parameter( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_write_parameter()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned short short_value;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Record '%s', channel %d, parameter '%s', value = %lu.",
			fname, mca->record->name, xia_dxp_mca->system_channel,
			parameter_name, (unsigned long) value));
	}

	short_value = (unsigned short) value;

	xia_status = dxp_set_one_dspsymbol( &(xia_dxp_mca->system_channel),
						parameter_name, &short_value );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot write the value %hu to DXP parameter '%s' "
		"for system channel %d of XIA DXP interface '%s'.  "
		"Error code = %d, error status = '%s'", short_value,
			parameter_name, xia_dxp_mca->system_channel,
			mca->record->name, xia_status,
			mxi_xia_xerxes_strerror( xia_status ) );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Write to parameter '%s' succeeded.",
			fname, parameter_name));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_write_parameter_to_all_channels( MX_MCA *mca,
			char *parameter_name,
			mx_uint32_type value,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_write_parameter_to_all_channels()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned short short_value;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s', parameter '%s', value = %lu.",
			fname, mca->record->name,
			parameter_name, (unsigned long) value));
	}

	short_value = (unsigned short) value;

#if 0
	/* The canonical method which does not work. */

	xia_status = dxp_set_dspsymbol( parameter_name, &short_value );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot write the value %hu to DXP parameter '%s' "
		"for all channels of XIA DXP interface '%s'.  ",
		"Error code = %d, error status = '%s'",
			short_value, parameter_name, mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}
#else
	/* A temporary kludge which does work. */

	{
		MX_RECORD *mca_record;
		MX_XIA_DXP_MCA *local_dxp_mca;
		unsigned long i;

		for ( i = 0; i < xia_xerxes->num_mcas; i++ ) {

			mca_record = xia_xerxes->mca_record_array[i];

			if ( mca_record == (MX_RECORD *) NULL ) {

				continue;    /* Skip this one. */
			}

			local_dxp_mca = (MX_XIA_DXP_MCA *)
					mca_record->record_type_struct;

			MX_DEBUG(-2,("%s: writing to system channel %d.",
				fname, local_dxp_mca->system_channel));

			xia_status = dxp_set_one_dspsymbol(
					&(local_dxp_mca->system_channel),
					parameter_name, &short_value );

			if ( xia_status != DXP_SUCCESS ) {
				return mx_error( MXE_DEVICE_ACTION_FAILED,fname,
			"Cannot write the value %hu to DXP parameter '%s' "
			"for system channel %d of XIA DXP interface '%s'.  "
			"Error code = %d, error status = '%s'", short_value,
				parameter_name, local_dxp_mca->system_channel,
				mca->record->name, xia_status,
				mxi_xia_xerxes_strerror( xia_status ) );
			}
		}
	}
#endif
	if ( debug_flag ) {
		MX_DEBUG(-2,
		("%s: Write parameter '%s' to all channels succeeded.",
			fname, parameter_name));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_start_run( MX_MCA *mca,
			int clear_flag,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_start_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned short gate, resume_flag;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s', clear_flag = %d.",
				fname, mca->record->name, clear_flag));
	}

	gate = 1;

	if ( clear_flag ) {
		resume_flag = 0;
	} else {
		resume_flag = 1;
	}

	xia_status = dxp_start_run( &gate, &resume_flag );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot start data acquisition for XIA DXP interface '%s'.  "
		"Error code = %d, error status = '%s'", mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	return MX_SUCCESSFUL_RESULT;
}


MX_EXPORT mx_status_type
mxi_xia_xerxes_stop_run( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_stop_run()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: Record '%s'.", fname, mca->record->name));
	}

	xia_status = dxp_stop_run();

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot stop data acquisition for XIA DXP interface '%s'.  "
		"Error code = %d, error status = '%s'", mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_xia_xerxes_read_spectrum( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_read_spectrum()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	int xia_status;
	unsigned long *array_ptr;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( xia_dxp_mca->use_mca_channel_array ) {
		array_ptr = mca->channel_array;
	} else {
		array_ptr = xia_dxp_mca->spectrum_array;
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: reading out %ld channels from MCA '%s'.",
			fname, mca->current_num_channels, mca->record->name));
	}

	xia_status = dxp_readout_detector_run(
				&(xia_dxp_mca->system_channel),
				xia_dxp_mca->parameter_array,
				xia_dxp_mca->baseline_array,
				array_ptr );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot read the spectrum for XIA MCA '%s'.  "
		"Error code = %d, error status = '%s'", mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s: readout from MCA '%s' complete.",
			fname, mca->record->name));
	}

	return MX_SUCCESSFUL_RESULT;
}

#define TWO_TO_THE_32ND_POWER	4294967296.0

MX_EXPORT mx_status_type
mxi_xia_xerxes_read_statistics( MX_MCA *mca,
			int debug_flag )
{
	const char fname[] = "mxi_xia_xerxes_read_statistics()";

	MX_XIA_DXP_MCA *xia_dxp_mca;
	MX_XIA_XERXES *xia_xerxes;
	unsigned long nevent;
	int xia_status;
	mx_status_type mx_status;

	mx_status = mxi_xia_xerxes_get_pointers( mca,
					&xia_dxp_mca, &xia_xerxes, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	xia_status = dxp_get_statistics( &(xia_dxp_mca->system_channel),
					&(mca->live_time),
					&(xia_dxp_mca->input_count_rate),
					&(xia_dxp_mca->output_count_rate),
					&nevent );

	if ( xia_status != DXP_SUCCESS ) {
		return mx_error( MXE_DEVICE_ACTION_FAILED, fname,
		"Cannot read statistics for XIA DXP interface '%s'.  "
		"Error code = %d, error status = '%s'", mca->record->name,
			xia_status, mxi_xia_xerxes_strerror( xia_status ) );
	}

	if ( debug_flag ) {
		MX_DEBUG(-2,(
"%s: Record '%s', channel %d, live_time = %g, icr = %g, ocr = %g, nevent = %lu",
			fname, mca->record->name, xia_dxp_mca->system_channel,
			mca->live_time, xia_dxp_mca->input_count_rate,
			xia_dxp_mca->output_count_rate, nevent ));
	}

	xia_dxp_mca->num_events = (mx_uint32_type) nevent;

	xia_dxp_mca->num_fast_peaks = (mx_uint32_type) mx_round(
				xia_dxp_mca->input_count_rate * mca->live_time);

	/* Hopefully, I can fix the following. */

	mca->real_time = mca->live_time;

	xia_dxp_mca->num_underflows = 0;
	xia_dxp_mca->num_overflows = 0;

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

const char *
mxi_xia_xerxes_strerror( int xia_status )
{
	static const char foo[] = "XerXes error messages are not yet here.";

	return &foo[0];
}


#endif /* HAVE_XIA_XERXES */
