/*
 * Name:    mx_mca.c
 *
 * Purpose: MX function library for multichannel analyzers.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_mca.h"

/*=======================================================================*/

/* This function is a utility function to consolidate all of the pointer
 * mangling that often has to happen at the beginning of an mx_mca_...
 * function.  The parameter 'calling_fname' is passed so that error messages
 * will appear with the name of the calling function.
 */

MX_EXPORT mx_status_type
mx_mca_get_pointers( MX_RECORD *mca_record,
			MX_MCA **mca,
			MX_MCA_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_mca_get_pointers()";

	if ( mca_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The mca_record pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( mca_record->mx_class != MXC_MULTICHANNEL_ANALYZER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"The record '%s' passed by '%s' is not an MCA.",
			mca_record->name, calling_fname );
	}

	if ( mca != (MX_MCA **) NULL ) {
		*mca = (MX_MCA *) (mca_record->record_class_struct);

		if ( *mca == (MX_MCA *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCA pointer for record '%s' passed by '%s' is NULL",
				mca_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_MCA_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_MCA_FUNCTION_LIST *)
				(mca_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_MCA_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_MCA_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				mca_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_mca_finish_record_initialization( MX_RECORD *mca_record )
{
	const char fname[] = "mx_mca_finish_record_initialization()";

	MX_MCA *mca;
	MX_RECORD_FIELD *channel_array_field;
	mx_status_type mx_status;

	if ( mca_record == ( MX_RECORD * ) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}
	mca = (MX_MCA *) mca_record->record_class_struct;

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MCA pointer for record '%s' is NULL.",
			mca_record->name );
	}

	/* Initialize some fields in the MX_MCA structure. */

	mca->start = 0;
	mca->stop = 0;
	mca->clear = 0;
	mca->busy = 0;

	/* Since 'channel_array' is a varying length field, but is not
	 * specified in the record description, we must fix up the
	 * dimension[0] value for the field by hand.
	 */

	mx_status = mx_find_record_field( mca_record, "channel_array",
						&channel_array_field );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	channel_array_field->dimension[0] = (long) mca->maximum_num_channels;

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mx_mca_start( MX_RECORD *mca_record )
{
	const char fname[] = "mx_mca_start()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *start_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	start_fn = function_list->start;

	if ( start_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"start function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	status = (*start_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_stop( MX_RECORD *mca_record )
{
	const char fname[] = "mx_mca_stop()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *stop_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	stop_fn = function_list->stop;

	if ( stop_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"stop function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	status = (*stop_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_read( MX_RECORD *mca_record,
		unsigned long *num_channels,
		unsigned long **channel_array )
{
	const char fname[] = "mx_mca_read()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *read_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	read_fn = function_list->read;

	if ( read_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	status = (*read_fn)( mca );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_channels != NULL ) {
		*num_channels = mca->current_num_channels;
	}

	if ( channel_array != NULL ) {
		*channel_array = mca->channel_array;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_mca_clear( MX_RECORD *mca_record )
{
	const char fname[] = "mx_mca_clear()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *clear_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	clear_fn = function_list->clear;

	if ( clear_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"clear function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	status = (*clear_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_is_busy( MX_RECORD *mca_record, int *busy )
{
	const char fname[] = "mx_mca_is_busy()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *busy_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	busy_fn = function_list->busy;

	if ( busy_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"busy function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	status = (*busy_fn)( mca );

	*busy = mca->busy;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_start_without_preset( MX_RECORD *mca_record )
{
	const char fname[] = "mx_mca_start_without_preset()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *start_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	start_fn = function_list->start;

	if ( start_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"start function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	mca->preset_type = MXF_MCA_PRESET_NONE;

	status = (*start_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_start_for_preset_live_time( MX_RECORD *mca_record,
					double preset_seconds )
{
	const char fname[] = "mx_mca_start_for_preset_live_time()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *start_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	start_fn = function_list->start;

	if ( start_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"start function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	mca->preset_type = MXF_MCA_PRESET_LIVE_TIME;

	mca->preset_live_time = preset_seconds;

	status = (*start_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_start_for_preset_real_time( MX_RECORD *mca_record,
					double preset_seconds )
{
	const char fname[] = "mx_mca_start_for_preset_real_time()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *start_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	start_fn = function_list->start;

	if ( start_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"start function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	mca->preset_type = MXF_MCA_PRESET_REAL_TIME;

	mca->preset_real_time = preset_seconds;

	status = (*start_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_roi( MX_RECORD *mca_record, unsigned long *roi )
{
	const char fname[] = "mx_mca_get_roi()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_parameter ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	if ( roi == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"roi pointer passed is NULL." );
	}
	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_parameter = function_list->get_parameter;

	if ( get_parameter == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_parameter function ptr for record '%s' is NULL.",
			mca_record->name);
	}

	mca->parameter_type = MXLV_MCA_ROI;

	status = (*get_parameter)( mca );

	roi[0] = mca->roi[0];
	roi[1] = mca->roi[1];

	return status;
}

MX_EXPORT mx_status_type
mx_mca_set_roi( MX_RECORD *mca_record, unsigned long *roi )
{
	const char fname[] = "mx_mca_set_roi()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *set_parameter ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	set_parameter = function_list->set_parameter;

	if ( set_parameter == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"set_roi function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	mca->parameter_type = MXLV_MCA_ROI;

	mca->roi[0] = roi[0];
	mca->roi[1] = roi[1];

	status = (*set_parameter)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_roi_integral( MX_RECORD *mca_record,
			unsigned long *roi_integral )
{
	const char fname[] = "mx_mca_get_roi_integral()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_parameter ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	if ( roi_integral == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"roi_integral pointer passed is NULL." );
	}

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_parameter = function_list->get_parameter;

	if ( get_parameter == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_parameter function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	mca->parameter_type = MXLV_MCA_ROI_INTEGRAL;

	status = (*get_parameter)( mca );

	*roi_integral = mca->roi_integral;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_num_channels( MX_RECORD *mca_record,
			unsigned long *num_channels )
{
	const char fname[] = "mx_mca_get_num_channels()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_parameter ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	if ( num_channels == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"num_channels pointer passed is NULL." );
	}

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_parameter = function_list->get_parameter;

	if ( get_parameter == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_parameter function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	mca->parameter_type = MXLV_MCA_CURRENT_NUM_CHANNELS;

	status = (*get_parameter)( mca );

	*num_channels = mca->current_num_channels;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_set_num_channels( MX_RECORD *mca_record,
			unsigned long num_channels )
{
	const char fname[] = "mx_mca_set_num_channels()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *set_parameter ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	set_parameter = function_list->set_parameter;

	if ( set_parameter == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"set_parameter function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	mca->parameter_type = MXLV_MCA_CURRENT_NUM_CHANNELS;

	mca->current_num_channels = num_channels;

	status = (*set_parameter)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_channel( MX_RECORD *mca_record,
			unsigned long channel_number,
			unsigned long *channel_value )
{
	const char fname[] = "mx_mca_get_channel()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *set_parameter ) ( MX_MCA * );
	mx_status_type ( *get_parameter ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	set_parameter = function_list->set_parameter;

	if ( set_parameter == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"set_parameter function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	get_parameter = function_list->get_parameter;

	if ( get_parameter == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_parameter function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	/* Set the channel number to be read. */

	mca->parameter_type = MXLV_MCA_CHANNEL_NUMBER;

	mca->channel_number = channel_number;

	status = (*set_parameter)( mca );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Now read back the channel value. */

	mca->parameter_type = MXLV_MCA_CHANNEL_VALUE;

	status = (*get_parameter)( mca );

	*channel_value = mca->channel_value;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_real_time( MX_RECORD *mca_record, double *real_time )
{
	const char fname[] = "mx_mca_get_real_time()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_parameter ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	if ( real_time == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"real_time pointer passed is NULL." );
	}

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_parameter = function_list->get_parameter;

	if ( get_parameter == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_parameter function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	mca->parameter_type = MXLV_MCA_REAL_TIME;

	status = (*get_parameter)( mca );

	*real_time = mca->real_time;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_live_time( MX_RECORD *mca_record, double *live_time )
{
	const char fname[] = "mx_mca_get_live_time()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_parameter ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", fname, mca_record->name));

	if ( live_time == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"live_time pointer passed is NULL." );
	}

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_parameter = function_list->get_parameter;

	if ( get_parameter == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_parameter function ptr for record '%s' is NULL.",
			mca_record->name );
	}

	mca->parameter_type = MXLV_MCA_LIVE_TIME;

	status = (*get_parameter)( mca );

	*live_time = mca->live_time;

	return status;
}

