/*
 * Name:     mx_syslg.c
 *
 * Purpose:  Handler for Unix style syslogd daemon support.
 *
 * Author:   William Lavender
 *
 *------------------------------------------------------------------------
 *
 * Copyright 2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mx_osdef.h"

#ifdef OS_UNIX
#include <syslog.h>
#endif

#include "mx_rec.h"
#include "mx_syslg.h"

#ifndef OS_UNIX

MX_EXPORT mx_status_type
mx_install_syslog_handler( char *ident_string,
				int log_number,
				int log_options )
{

	mx_info(
    "Unix style syslog support is not available for this operating system." );

	return MX_SUCCESSFUL_RESULT;
}

#else /* OS_UNIX */

static void mx_syslog_handler( int level, char *message )
{
	syslog( level, message );
}

static void mx_syslog_info_handler( char *message )
{
	mx_syslog_handler( LOG_INFO, message );
}

static void mx_syslog_warning_handler( char *message )
{
	mx_syslog_handler( LOG_WARNING, message );
}

static void mx_syslog_error_handler( char *message )
{
	mx_syslog_handler( LOG_ERR, message );
}

static void mx_syslog_debug_handler( char *message )
{
	mx_syslog_handler( LOG_DEBUG, message );
}

MX_EXPORT mx_status_type
mx_install_syslog_handler( char *ident_string,
				int log_number,
				int log_options )
{
	const char fname[] = "mx_install_syslog_handler()";

	int syslog_facility, syslog_options;

	syslog_facility = -1;
	syslog_options = 0;

	switch ( log_number ) {
	case 0:
		syslog_facility = LOG_LOCAL0;
		break;
	case 1:
		syslog_facility = LOG_LOCAL1;
		break;
	case 2:
		syslog_facility = LOG_LOCAL2;
		break;
	case 3:
		syslog_facility = LOG_LOCAL3;
		break;
	case 4:
		syslog_facility = LOG_LOCAL4;
		break;
	case 5:
		syslog_facility = LOG_LOCAL5;
		break;
	case 6:
		syslog_facility = LOG_LOCAL6;
		break;
	case 7:
		syslog_facility = LOG_LOCAL7;
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Illegal syslog log number %d.  The allowed values are from 0 to 7.",
			log_number );
		break;
	}

#ifdef LOG_PERROR
	if ( log_options & MXF_SYSLOG_USE_STDERR ) {
		syslog_options = LOG_PERROR | LOG_PID;
	} else {
		syslog_options = LOG_PID;
	}
#else
	syslog_options = LOG_PID;
#endif

	openlog( ident_string, syslog_options, syslog_facility );

	mx_set_info_output_function( mx_syslog_info_handler );
	mx_set_warning_output_function( mx_syslog_warning_handler );
	mx_set_error_output_function( mx_syslog_error_handler );
	mx_set_debug_output_function( mx_syslog_debug_handler );

	return MX_SUCCESSFUL_RESULT;
}

#endif /* OS_UNIX */

