/*
 * Name:    mx_timer.c
 *
 * Purpose: MX function library of generic timer operations.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999, 2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_timer.h"

/*=======================================================================*/

/* This function is a utility function to consolidate all of the pointer
 * mangling that often has to happen at the beginning of an mx_timer_...
 * function.  The parameter 'calling_fname' is passed so that error messages
 * will appear with the name of the calling function.
 */

MX_EXPORT mx_status_type
mx_timer_get_pointers( MX_RECORD *timer_record,
			MX_TIMER **timer,
			MX_TIMER_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_timer_get_pointers()";

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The timer_record pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( timer_record->mx_class != MXC_TIMER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"The record '%s' passed by '%s' is not a timer.",
			timer_record->name, calling_fname );
	}

	if ( timer != (MX_TIMER **) NULL ) {
		*timer = (MX_TIMER *) (timer_record->record_class_struct);

		if ( *timer == (MX_TIMER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_TIMER pointer for record '%s' passed by '%s' is NULL",
				timer_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_TIMER_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_TIMER_FUNCTION_LIST *)
				(timer_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_TIMER_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_TIMER_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				timer_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mx_timer_finish_record_initialization( MX_RECORD *timer_record )
{
	const char fname[] = "mx_timer_finish_record_initialization()";

	MX_TIMER *timer;

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	timer = (MX_TIMER *)(timer_record->record_class_struct);

	if ( timer == (MX_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TIMER pointer for record '%s' is NULL.",
			timer_record->name );
	}

	timer->value = 0.0;
	timer->busy = 0;
	timer->stop = 0;
	timer->clear = 0;
	timer->last_measurement_time = -1.0;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_timer_is_busy( MX_RECORD *timer_record, int *busy )
{
	const char fname[] = "mx_timer_is_busy()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_is_busy_fn ) ( MX_TIMER * );
	mx_status_type status;

	status = mx_timer_get_pointers( timer_record,
					&timer, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	timer_is_busy_fn = function_list->timer_is_busy;

	if ( timer_is_busy_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"timer_is_busy function ptr for timer record '%s' is NULL.",
			timer_record->name );
	}

	status = (*timer_is_busy_fn)( timer );

	*busy = timer->busy;

	return status;
}

MX_EXPORT mx_status_type
mx_timer_start( MX_RECORD *timer_record, double seconds )
{
	const char fname[] = "mx_timer_start()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_start ) ( MX_TIMER * );
	mx_status_type status;

	status = mx_timer_get_pointers( timer_record,
					&timer, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	timer_start = function_list->start;

	if ( timer_start == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"start function ptr for timer record '%s' is NULL.",
			timer_record->name );
	}

	timer->value = seconds;

	status = (*timer_start)( timer );

	timer->last_measurement_time = seconds;

	return status;
}

MX_EXPORT mx_status_type
mx_timer_stop( MX_RECORD *timer_record, double *seconds_left )
{
	const char fname[] = "mx_timer_stop()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_stop ) ( MX_TIMER * );
	mx_status_type status;

	status = mx_timer_get_pointers( timer_record,
					&timer, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	timer_stop = function_list->stop;

	if ( timer_stop == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"stop function ptr for timer record '%s' is NULL.",
			timer_record->name );
	}

	status = (*timer_stop)( timer );

	*seconds_left = timer->value;

	timer->last_measurement_time = -1.0;

	return status;
}

MX_EXPORT mx_status_type
mx_timer_clear( MX_RECORD *timer_record )
{
	const char fname[] = "mx_timer_clear()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_clear ) ( MX_TIMER * );
	mx_status_type status;

	status = mx_timer_get_pointers( timer_record,
					&timer, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	timer_clear = function_list->clear;

	if ( timer_clear == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"clear function ptr for timer record '%s' is NULL.",
			timer_record->name );
	}

	status = (*timer_clear)( timer );

	timer->last_measurement_time = -1.0;

	return status;
}

MX_EXPORT mx_status_type
mx_timer_read( MX_RECORD *timer_record, double *seconds )
{
	const char fname[] = "mx_timer_read()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_read ) ( MX_TIMER * );
	mx_status_type status;

	status = mx_timer_get_pointers( timer_record,
					&timer, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	timer_read = function_list->read;

	if ( timer_read == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for timer record '%s' is NULL.",
			timer_record->name );
	}

	/* All is well, so invoke the function. */

	status = (*timer_read)( timer );

	*seconds = timer->value;

	return status;
}

MX_EXPORT mx_status_type
mx_timer_get_mode( MX_RECORD *timer_record, int *mode )
{
	const char fname[] = "mx_timer_get_mode()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_get_mode ) ( MX_TIMER * );
	mx_status_type status;

	status = mx_timer_get_pointers( timer_record,
					&timer, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	timer_get_mode = function_list->get_mode;

	if ( timer_get_mode == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_mode function ptr for timer record '%s' is NULL.",
			timer_record->name );
	}

	status = (*timer_get_mode)( timer );

	if ( status.code != MXE_SUCCESS )
		return status;

	*mode = timer->mode;

	return status;
}

MX_EXPORT mx_status_type
mx_timer_set_mode( MX_RECORD *timer_record, int mode )
{
	const char fname[] = "mx_timer_set_mode()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_set_mode ) ( MX_TIMER * );
	mx_status_type status;

	status = mx_timer_get_pointers( timer_record,
					&timer, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	timer_set_mode = function_list->set_mode;

	if ( timer_set_mode == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"set_mode function ptr for timer record '%s' is NULL.",
			timer_record->name );
	}

	timer->mode = mode;

	status = (*timer_set_mode)( timer );

	timer->last_measurement_time = -1.0;

	return status;
}

MX_EXPORT mx_status_type
mx_timer_set_modes_of_associated_counters( MX_RECORD *timer_record )
{
	const char fname[] = "mx_timer_set_modes_of_associated_counters()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *set_modes_of_associated_counters ) ( MX_TIMER * );
	mx_status_type status;

	status = mx_timer_get_pointers( timer_record,
					&timer, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	set_modes_of_associated_counters =
			function_list->set_modes_of_associated_counters;

	/* Do nothing if the driver does not define this function. */

	if ( set_modes_of_associated_counters == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	status = (*set_modes_of_associated_counters)( timer );

	return status;
}

