/*
 * Name:    ms_prmtr.c
 *
 * Purpose: Functions used to process MX motor record events.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2001 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_rec.h"
#include "mx_motor.h"
#include "mx_sock.h"

#include "ms_serv.h"
#include "ms_prmtr.h"

mx_status_type
mxsrv_setup_motor_process_functions( MX_RECORD *record_list_head )
{
	const char fname[] = "mxsrv_setup_motor_process_functions()";

	MX_RECORD *current_record;
	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD *record_field_array;
	long i;

	MX_DEBUG( 2,("%s invoked.", fname));

	/* The following is done just in case we were called with
	 * a record that isn't really the list head.
	 */

	record_list_head = record_list_head->list_head;

	/* Go through the record list looking for motor records. */

	current_record = record_list_head->next_record;

	while ( current_record != record_list_head ) {
		if ( (current_record->mx_superclass == MXR_DEVICE)
		  && (current_record->mx_class == MXC_MOTOR) ) {

			record_field_array
				= current_record->record_field_array;

			for ( i = 0;
				i < current_record->num_record_fields;
				i++ )
			{
				record_field = &record_field_array[i];

				switch( record_field->label_value ) {
				case MXLV_MTR_BUSY:
				case MXLV_MTR_DESTINATION:
				case MXLV_MTR_POSITION:
				case MXLV_MTR_SET_POSITION:
				case MXLV_MTR_SOFT_ABORT:
				case MXLV_MTR_IMMEDIATE_ABORT:
				case MXLV_MTR_POSITIVE_LIMIT_HIT:
				case MXLV_MTR_NEGATIVE_LIMIT_HIT:
				case MXLV_MTR_CONSTANT_VELOCITY_MOVE:
				case MXLV_MTR_SPEED:
				case MXLV_MTR_BASE_SPEED:
				case MXLV_MTR_MAXIMUM_SPEED:
				case MXLV_MTR_ACCELERATION_PARAMETERS:
				case MXLV_MTR_ACCELERATION_TIME:
				case MXLV_MTR_ACCELERATION_DISTANCE:
				case MXLV_MTR_SPEED_CHOICE_PARAMETERS:
				case MXLV_MTR_SAVE_SPEED:
				case MXLV_MTR_RESTORE_SPEED:
				case MXLV_MTR_SYNCHRONOUS_MOTION_MODE:
				case MXLV_MTR_COMPUTE_EXTENDED_SCAN_RANGE:
				case MXLV_MTR_COMPUTE_PSEUDOMOTOR_POSITION:
				case MXLV_MTR_COMPUTE_REAL_POSITION:
					record_field->process_function
					    = mxsrv_motor_process_function;
					break;
				default:
					break;
				}
			}
		}
		current_record = current_record->next_record;
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_motor_process_function( void *record_ptr,
			void *record_field_ptr, int operation )
{
	const char fname[] = "mxsrv_motor_process_function()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *motor_function_list;
	char operation_string[30];
	long label_value;
	double position1, position2, time_for_move;
	mx_status_type status;

	double position, speed, acceleration_time, acceleration_distance;
	double acceleration_params[MX_MOTOR_NUM_ACCELERATION_PARAMS];
	double start_position, end_position;
	double real_start_position, real_end_position;
	double pseudomotor_position, real_position;
	int busy, limit_hit, flags;

	record = (MX_RECORD *) record_ptr;
	record_field = (MX_RECORD_FIELD *) record_field_ptr;

	switch ( operation ) {
	case MX_PROCESS_GET:
		strcpy( operation_string, "(MX_PROCESS_GET)" );
		break;
	case MX_PROCESS_PUT:
		strcpy( operation_string, "(MX_PROCESS_PUT)" );
		break;
	default:
		sprintf( operation_string, "(Operation %d)", operation );
		break;
	}

	label_value = record_field->label_value;

	MX_DEBUG(1,("%s: '%s.%s' %s (%ld)",
		fname, record->name, record_field->name,
		operation_string, label_value));

	/*
	 * A label value < 0 tells us that we do not need to do any
	 * further MX record processing. 
	 */

	if ( label_value < 0 ) {
		return MX_SUCCESSFUL_RESULT;
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	motor_function_list = (MX_MOTOR_FUNCTION_LIST *)
				(record->class_specific_function_list);

	status = MX_SUCCESSFUL_RESULT;

	switch( operation ) {
	case MX_PROCESS_GET:
		switch( label_value ) {
		case MXLV_MTR_POSITION:
			status = mx_motor_get_position( record, &position );
			break;
		case MXLV_MTR_BUSY:
			status = mx_motor_is_busy( record, &busy );
			break;
		case MXLV_MTR_NEGATIVE_LIMIT_HIT:
			status = mx_motor_negative_limit_hit(
						record, &limit_hit );
			break;
		case MXLV_MTR_POSITIVE_LIMIT_HIT:
			status = mx_motor_positive_limit_hit(
						record, &limit_hit );
			break;
		case MXLV_MTR_SPEED:
			status = mx_motor_get_speed( record, &speed );

			break;
		case MXLV_MTR_BASE_SPEED:
			status = mx_motor_get_base_speed( record, &speed );

			break;
		case MXLV_MTR_MAXIMUM_SPEED:
			status = mx_motor_get_base_speed( record, &speed );

			break;
		case MXLV_MTR_SYNCHRONOUS_MOTION_MODE:
			status = mx_motor_get_synchronous_motion_mode(
							record, NULL );
			break;
		case MXLV_MTR_ACCELERATION_PARAMETERS:
			status = mx_motor_get_acceleration_parameters(
						record, acceleration_params );

			break;
		case MXLV_MTR_ACCELERATION_TIME:
			status = mx_motor_get_acceleration_time(
						record, &acceleration_time );
			break;
		case MXLV_MTR_ACCELERATION_DISTANCE:
			status = mx_motor_get_acceleration_distance(
					record, &acceleration_distance );
			break;
		case MXLV_MTR_COMPUTE_EXTENDED_SCAN_RANGE:
			start_position = motor->compute_extended_scan_range[0];
			end_position = motor->compute_extended_scan_range[1];

			status = mx_motor_compute_extended_scan_range( record,
				start_position, end_position,
				&real_start_position, &real_end_position );
			break;
		case MXLV_MTR_COMPUTE_PSEUDOMOTOR_POSITION:
			real_position = motor->compute_pseudomotor_position[0];

			status =
		mx_motor_compute_pseudomotor_position_from_real_position(
			record, real_position, &pseudomotor_position, TRUE );

			break;
		case MXLV_MTR_COMPUTE_REAL_POSITION:
			pseudomotor_position = motor->compute_real_position[0];

			status =
		mx_motor_compute_real_position_from_pseudomotor_position(
			record, pseudomotor_position, &real_position, TRUE );

			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_GET label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	case MX_PROCESS_PUT:
		switch( label_value ) {
		case MXLV_MTR_DESTINATION:
			flags = MXF_MTR_NOWAIT | MXF_MTR_IGNORE_KEYBOARD
					| MXF_MTR_IGNORE_BACKLASH;

			status = mx_motor_move_absolute( record,
					motor->destination, flags );

			break;
		case MXLV_MTR_SET_POSITION:
			status = mx_motor_set_position( record,
						motor->set_position );

			break;
		case MXLV_MTR_SOFT_ABORT:
			status = mx_motor_soft_abort( record );

			/* Set the soft abort flag back to zero. */

			motor->soft_abort = 0;
			break;
		case MXLV_MTR_IMMEDIATE_ABORT:
			status = mx_motor_immediate_abort( record );

			/* Set the immediate abort flag back to zero. */

			motor->immediate_abort = 0;
			break;
		case MXLV_MTR_CONSTANT_VELOCITY_MOVE:
			status = mx_motor_constant_velocity_move( record,
					motor->constant_velocity_move );
			break;
		case MXLV_MTR_SPEED:
			status = mx_motor_set_speed( record, motor->speed );

			break;
		case MXLV_MTR_BASE_SPEED:
			status = mx_motor_set_base_speed( record,
							motor->base_speed );

			break;
		case MXLV_MTR_MAXIMUM_SPEED:
			status = mx_motor_set_base_speed( record,
							motor->maximum_speed );

			break;
		case MXLV_MTR_ACCELERATION_PARAMETERS:
			status = mx_motor_set_acceleration_parameters( record,
					motor->acceleration_parameters );

			break;
		case MXLV_MTR_SPEED_CHOICE_PARAMETERS:
			position1 = motor->speed_choice_parameters[0];
			position2 = motor->speed_choice_parameters[1];
			time_for_move = motor->speed_choice_parameters[2];

			status = mx_motor_set_speed_between_positions( record,
					position1, position2, time_for_move );
			break;
		case MXLV_MTR_SAVE_SPEED:
			status = mx_motor_save_speed( record );
			break;
		case MXLV_MTR_RESTORE_SPEED:
			status = mx_motor_restore_speed( record );
			break;
		case MXLV_MTR_SYNCHRONOUS_MOTION_MODE:
			status = mx_motor_set_synchronous_motion_mode( record,
					motor->synchronous_motion_mode );
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_PUT label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown operation code = %d", operation );
		break;
	}

	return status;
}

